""" 
This file contains the API for the PilotData Framework.
"""
    
class PilotDataDescription(dict):
    """ PilotDataDescription.
        {
            'service_url': "ssh://localhost/tmp/pilotstore/",
            'size':100,
            
            # Affinity
            'affinity_datacenter_label',    # pilot stores sharing the same label are located in the same data center          
            'affinity_machine_label',       # pilot stores sharing the same label are located on the same machine                           
        }    
    """
    
    def __init__(self):
        pass
    

    def __setattr__(self, attr, value):
        self[attr]=value
        
    
    def __getattr__(self, attr):
        return self[attr]



class PilotData(object):
    """ PilotData.  """

    # Class members
    __slots__ = (
        'id',           # Reference to this PJ
        'description',  # Description of PilotData
        'context',      # SAGA context
        'resource_url', # Resource  URL       
        'state',        # State of the PilotData
        'state_detail', # Adaptor specific state of the PilotData       
    )

    def cancel(self):        
        """ Cancel PilotData 

            Keyword arguments:
            None
        """
        pass
    
    
    def get_state(self):
        pass
    
       


class PilotDataService(object):
    """ PilotDataService (PDS).   """

    # Class members
    __slots__ = (
        'id',           # Reference to this PJS
        'state',       # Status of the PJS
        'pilot_stores'    # List of PJs under this PJS
    )

    def __init__(self, pss_id=None):
        """ Create a PilotDataService

            Keyword arguments:
            pss_id -- restore from pss_id
        """
        pass

    def create_pilot(self, pilot_data_description):
        """ Create a PilotData 

            Keyword arguments:
            pilot_data_description -- PilotData Description            
            
            Return value:
            A PilotData handle
        """
        pass

    def list_pilots(self):
        """ List all PDs of PDS """
        pass

    def cancel(self):
        """ Cancel the PilotDataService.
            
            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        pass


#
# PilotDataService
# 
class DataUnitService(object):
    """ PilotDataService (PDS).   
    """
   
    def __init__(self, pds_id=None):
        """ Create a DataUnitService.

            Keyword arguments:
            pds_id -- Reconnect to an existing DataUnitService 
        """
        pass

    def add_pilot_data_service(self, pss):
        """ Add a PilotDataService 

            Keyword arguments:
            pss -- The PilotDataService to add.

            Return:
            Result
        """
        pass

    def remove_pilot_data_service(self, pss):
        """ Remove a PilotDataService 

            
            Keyword arguments:
            pss -- The PilotDataService to remove 
            Return:
            Result
        """
        pass
    
    
    def list_pilot_data(self):
        """ List all PDs of PDS """
        pass
    

    def submit_pilot_data_set(self, data_unit_description):
        """ Create Pilot Data object and schedule it to a Pilot Data """
        pass
    
    
    def cancel(self):
        """ Cancel the PDS.            
            
            Keyword arguments:
            None

            Return:
            Result
        """
        pass

    
    def get_state(self):
        pass
    
    
    def get_id(self):
        pass

#
# DataUnitDescription
# 
class DataUnitDescription(dict):
    """ DataUnitDescription.
        {
            'file_urls': [file1, file2, file3]        
        } 
        
        Currently, no directories supported
    """

    def __init__(self):
        pass

    def __setattr__(self, attr, value):
        self[attr]=value
    
    def __getattr__(self, attr):
        return self[attr]
    
    
#
# DataUnit 
# 
class DataUnit(object):
    """ DataUnit. Holds a set of files.        
    """

    # Class members
    __slots__ = (
        'id',               # Reference 
        'description',      # Description
        'state',            # State
        'file_list'        # files managed by PilotData object
    )

    def cancel(self):
        """ Cancel the PD. """
        pass
    
    def add_data_unit(self, data_unit):
        pass
    
    def remove_data_unit(self, data_unit):
        pass
    
    def list_data_units(self):
        pass
        
    def get_state(self):
        """
            get current state of Pilot Data:
                New => Initialized
                Pending => Files are synchronized with a pilot store
                Running => PD is in sync with all replicas
                Done => Terminated
        """
        pass
    
    
    def wait(self):
        """ Wait until in Running state 
            (or Failed state)
        """
        pass
    
    def export(self, target_directory):
        """ copies content of PD to a directory on the local machine"""
        pass    
        



import compute.api
import data.api



class ComputeDataDescription(compute.api.ComputeUnitDescription):
    """ ComputeDataDescription.
        {
            'executable': <path to executable>,  
            'arguments': <arguments>,           # Arguments 
            'environment': <environment>,          # environment variables 
            
            # Working directory
            # A pilot data url will be translated to the local pilot store url, e.g.
            #     pilotdata://localhost/c2fafeae-03a9-11e1-9934-109addae22a3' 
            #     =>           
            #     ssh://localhost/tmp/pilotdata/c2fafeae-03a9-11e1-9934-109addae22a3
            #
            #     i.e. job is executed in /tmp/pilotdata/c2fafeae-03a9-11e1-9934-109addae22a3
            #     where file can be accessed using relative paths
            'working_directory': <working directory>,  
            
            # I/O
            'input': <stdin>,
            'error': <sterr>,
            'output': <stdout>,
                        
            # Parallelism
            'number_of_processes': <Total number of processes to start>,
            'processes_per_host':  <Nr of processes per host>,
            'threads_per_process': <Nr of threads to start per process>,
            'total_core_count':    <Total number of cores requested>,
            'spmd_variation':      <Type and startup mechanism>,
            
            # Requirements
            'candidate_hosts': [<url>, ...],
            'cpu_architecture': <architecture, e.g. x86_64>,
            'total_physical_memory': <memory, e.g. 2000>,
            'operating_system_type': <os, e.g. Linux>,
            'total_cpu_time': <cpu time>,
            'wall_time_limit': <walltime in sec, e.g. 600 (sec)>,            
            
            # Data - input/output data flow for ComputeUnit
            'input_data': [<data unit url>, ... ],      
            'output_data': [<data unit url>, ... ]
        }
    """
    
    def __setattr__(self, attr, value):
        self[attr]=value
    
    def __getattr__(self, attr):
        return self[attr]

    


class ComputeDataService(compute.api.ComputeUnitService, data.api.PilotDataService):
    """ ComputeDataService.
    
        The ComputeDataService is the application's interface to submit 
        ComputeUnits and PilotData/DataUnit to the Pilot-Manager 
        in the P* Model.        
    """
   

    def __init__(self, wds_id=None):
        """ Create a Compute Data Service object.

            Keyword arguments:
            wds_id -- Reconnect to an existing WDS (optional).
        """
        pass   




"""
creates bootstrap script bigjob-bootstrap.py that installs bigjob Python package
""" 
import virtualenv, textwrap

def create_bigjob_bootstrap_script():
    output = virtualenv.create_bootstrap_script(textwrap.dedent("""
    import os, subprocess
    def after_install(options, home_dir):
        etc = join(home_dir, 'etc')
        if not os.path.exists(etc):
            os.makedirs(etc)         
        subprocess.call([join(home_dir, 'bin', 'easy_install'),
                     'bigjob'])    
    """))
    return output

if __name__ == "__main__" :
    output = create_bigjob_bootstrap_script()
    f = open('bigjob-bootstrap.py', 'w').write(output)


""" BigJob API

This Module contains the API of the BigJob framework

All BigJob implementation must be derived from this API.

"""

class bigjob(object):
       
    def __init__(self, database_host):  
        pass
    
    def start_pilot_job(self, 
                 lrms_url, 
                 bigjob_agent_executable,
                 number_nodes,
                 queue,
                 project,
                 working_directory,
                 userproxy,
                 walltime,
                 processes_per_node=1):
        pass
         
    def get_state(self):        
        pass
    
    def get_state_detail(self): 
        pass
    
    def cancel(self):        
        pass
    
    def wait(self):        
        pass
                    
                    
class subjob(object):
    
    def __init__(self, database_host=None):
        pass
    
    def submit_job(self, pilot_url, jd):
        pass
    
    def get_state(self):        
        pass
    
    def cancel(self):
        pass



import logging
import saga
import json
import pdb

from pilot import *
from bigjob import logger

class AdvertCoordinationAdaptor:
    """
        BigData persists its data in a central data space, e.g. the Advert service
        to facilitate distributed coordination:
        
        advert://advert.cct.lsu.edu/pilot/3d0d5960-296d-11e1-8896-00264a13ca4c/data/ => namespace for pilot data
        
        advert://advert.cct.lsu.edu/pilot/3d0d5960-296d-11e1-8896-00264a13ca4c/data/pds => pilot data service
        advert://advert.cct.lsu.edu/pilot/3d0d5960-296d-11e1-8896-00264a13ca4c/data/pds/pilot-data-description  => pilot data description
        ...
        
        
        advert://advert.cct.lsu.edu/pilot/3d0d5960-296d-11e1-8896-00264a13ca4c/data/pds/ => pilot store service
        advert://advert.cct.lsu.edu/pilot/3d0d5960-296d-11e1-8896-00264a13ca4c/data/pds/pilot-data-description => pilot data description
    
        This class is stateless - the application's base_url needs to be passed into every method.    
    """
    BASE_URL="advert://localhost/"
    BASE_URL_QUERY_STRING="?dbtype=sqlite3"
    
    PILOT_PATH="pilot"
    PILOT_DATA_PATH=PILOT_PATH
    PILOT_DATA_SERVICE_PATH=PILOT_DATA_PATH+"/pds"
    DATA_UNIT_SERVICE_PATH=PILOT_DATA_PATH+"/dus"
    COMPUTE_DATA_SERVICE_PATH = PILOT_DATA_PATH + "/cds"

    
    ###########################################################################
    # Construct a base url for an application
    
    @classmethod
    def get_base_url(cls, application_id):
        surl = saga.url(cls.BASE_URL)
        base_url = surl.scheme + "://" + surl.host + "/" + application_id + "/"
        logging.debug(base_url)
        return base_url
    
    ###########################################################################
    # Pilot Store Service related methods
    
    @classmethod  
    def add_pds(cls, application_url, pds):
        pds_url_no_dbtype = cls.get_pds_url(application_url, pds.id)
        pds_url = cls.__get_url(pds_url_no_dbtype)
        logger.debug("Create PDS directory at %s"%pds_url)
        saga.advert.directory(pds_url, saga.advert.Create | 
                                       saga.advert.CreateParents | 
                                       saga.advert.ReadWrite)
        return pds_url_no_dbtype
    
    
    @classmethod
    def delete_pds(cls, pds_url):
        pds_url = cls.__get_url(pds_url)
        pds_dir = saga.advert.directory(saga.url(pds_url), 
                                        saga.advert.Create | 
                                        saga.advert.CreateParents | 
                                        saga.advert.ReadWrite)
        pds_dir.remove(pds_url, saga.name_space.Recursive)  
    
    ###########################################################################
    # Pilot Data related methods
    
    @classmethod
    def add_pd(cls, pds_url, pd):
        pds_url = cls.__remove_dbtype(pds_url)        
        pd_url =pds_url+"/" + pd.id
        pd_description_url = cls.__get_url(pd_url + "/description")
        logger.debug("PDS URL: %s, PD Description URL: %s"%(pds_url, pd_description_url))
        # directory is recursively created
        pd_desc_entry = saga.advert.entry(saga.url(pd_description_url),
                                           saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        logger.debug("initialized advert entry for pds: " + pd_description_url)
        pd_desc_entry.store_string(json.dumps(pd.data_unit_description))
        return pd_url
    
    @classmethod
    def update_pd(cls, pd):
        if len(pd.data_units) > 0:
            du_urls = [i.url for i in pd.data_units.values()]
            cls.__store_entry(cls.__remove_dbtype(pd.url)+"/data-units", du_urls)
        cls.__store_entry(cls.__remove_dbtype(pd.url)+"/pilot-data", pd.to_dict())
    
    
    @classmethod
    def get_pd(cls, pds_url):
        logger.debug("GET PD: " + pds_url)     
        pd_dict={}        
        #pd_dict["pilot_data" ]=  cls.__retrieve_entry(cls.__remove_dbtype(pds_url)+"/pilot-data")
        pd_dict["pilot_data"] = cls.__retrieve_entry(cls.__remove_dbtype(pds_url)+"/pilot-data") 
        return pd_dict
        
    
    @classmethod
    def list_pd(cls, pds_url):
        """ return a list of urls to pd managed by the PDS """
        pds_url = cls.__get_url(pds_url)
        logger.debug("List PD at %s"%pds_url)
        pds_dir = saga.advert.directory(pds_url, saga.advert.Create | 
                                       saga.advert.CreateParents | 
                                       saga.advert.ReadWrite)
        
        pd_list = pds_dir.list()
        pd_full_urls = []
        for i in pd_list:
            pd_full_urls.append(pds_url + "/" + i)   
        return pd_full_urls
    
    @classmethod
    def delete_pd(cls, pds_url):
        pds_url = cls.__get_url(pds_url)
        pd_dir = saga.advert.directory(saga.url(pds_url), 
                                        saga.advert.Create | 
                                        saga.advert.CreateParents | 
                                        saga.advert.ReadWrite)
        pd_dir.remove(pds_url, saga.name_space.Recursive)  
    
        
    ###########################################################################
    # Compute Data Service related methods
    @classmethod  
    def add_cds(cls, application_url, cds):
        cds_url_no_dbtype = cls.get_cds_url(application_url, cds.id)
        cds_url = cls.__get_url(cds_url_no_dbtype)
        logger.debug("Create CDS directory at %s"%cds_url)
        saga.advert.directory(cds_url, saga.advert.Create | 
                                       saga.advert.CreateParents | 
                                       saga.advert.ReadWrite)
        return cds_url_no_dbtype
    
    @classmethod  
    def update_cds(cls, cds_url, cds):
        
        # Storage and Compute Resources
        pds_urls = [cls.__remove_dbtype(i.url) for i in cds.pilot_data_services]
        cls.__store_entry(cls.__remove_dbtype(cds_url)+"/pds/", pds_urls)
        
        pjs_urls = [i.url for i in cds.pilot_job_services]
        cls.__store_entry(cls.__remove_dbtype(cds_url)+"/cds/", pjs_urls)
        
        # currently managed PDs and WUs
        pd_urls = [i.url for i in cds.data_units.values()]
        cls.__store_entry(cls.__remove_dbtype(cds_url)+"/du/", pd_urls)
        
        wu_urls = [i.url for i in cds.compute_units.values()]
        cls.__store_entry(cls.__remove_dbtype(cds_url)+"/cu/", wu_urls)
            
        
    @classmethod
    def delete_cds(cls, cds_url):
        cds_url = cls.__get_url(cls.__remove_dbtype(cds_url))
        cds_dir = saga.advert.directory(saga.url(cds_url), 
                                        saga.advert.Create | 
                                        saga.advert.CreateParents | 
                                        saga.advert.ReadWrite)
        # cds_dir.remove(cds_url, saga.name_space.Recursive)
    
    
    
        
    ###########################################################################
    #  Data Unit related methods
    @classmethod
    def add_du(cls, dus_url, du):
        du_url = cls.__remove_dbtype(dus_url)  +  "/" + du.id     
        du_url = cls.__get_url(du_url)
        # directory is recursively created
        #saga.advert.directory(saga.url(du_url),
        #                                   saga.advert.Create | saga.advert.CreateParents | saga.advert.ReadWrite)
        #logger.debug("initialized advert entry for dus: " + du_url)
        return du_url

    
    @classmethod
    def get_du(cls, du_url):
        logger.debug("**** GET PD: " + du_url)
        du_dict={}        
        du_dict["data_unit_description" ]=  cls.__retrieve_entry(cls.__remove_dbtype(du_url)+"/description")
        du_dict["state"] = cls.__retrieve_entry(cls.__remove_dbtype(du_url)+"/state") 
        du_dict["data_units"] = cls.__retrieve_entry(cls.__remove_dbtype(du_url)+"/data-units")
        du_dict["pilot_data"] =  cls.__retrieve_entry(cls.__remove_dbtype(du_url)+"/pilot-data")    
        logger.debug("Open pilot data at: " + du_url + " State: " + str(du_dict))     
        return du_dict
    
     
    @classmethod  
    def update_du(cls, du):
        logger.debug("**** Update pilot data at: " + du.url)
        cls.__store_entry(cls.__remove_dbtype(du.url)+"/description", du.data_unit_description)
        cls.__store_entry(cls.__remove_dbtype(du.url)+"/state", du.state)
                
        du_urls = [i.url for i in du.pilot_data]
        cls.__store_entry(cls.__remove_dbtype(du.url)+"/pilot-data", du_urls)
                
        du_dict_list = [i.to_dict() for i in du.data_unit_items]
        cls.__store_entry(cls.__remove_dbtype(du.url)+"/data-units", du_dict_list)
        
       
    @classmethod
    def list_du(cls, dus_url):
        """ return a list of urls to du managed by the PDS """
        dus_url = cls.__get_url(dus_url)
        logger.debug("List PDS at %s"%dus_url)
        dus_dir = saga.advert.directory(dus_url, saga.advert.Create | 
                                       saga.advert.CreateParents | 
                                       saga.advert.ReadWrite)
        
        du_list = dus_dir.list()
        du_full_urls = []
        for i in du_list:
            du_full_urls.append(dus_url + "/" + i)   
        return du_full_urls
    
    
    @classmethod
    def delete_du(cls, du_url):
        du_url = cls.__get_url(du_url)
        du_dir = saga.advert.directory(saga.url(du_url), 
                                        saga.advert.Create | 
                                        saga.advert.CreateParents | 
                                        saga.advert.ReadWrite)
        du_dir.remove(du_url, saga.name_space.Recursive)  
    
    
    
    ###########################################################################
    # URL Tweaking
    
    @classmethod
    def get_pds_url(cls, application_url, pds_id):
        pds_url = application_url+AdvertCoordinationAdaptor.PILOT_DATA_SERVICE_PATH+"/"+pds_id        
        logger.debug("PDS URL: %s"%(pds_url))
        return pds_url
    
    @classmethod
    def get_cds_url(cls, application_url, cds_id):
        cds_url = application_url+AdvertCoordinationAdaptor.COMPUTE_DATA_SERVICE_PATH+"/"+cds_id        
        logger.debug("CDS URL: %s"%(cds_url))
        return cds_url
    
    ###########################################################################
    # internal methods
    
    @classmethod
    def __get_url(cls, url):
        """ appends advert querystring for dbtype to url """
        url = url + AdvertCoordinationAdaptor.BASE_URL_QUERY_STRING
        return url
    
    @classmethod
    def __remove_dbtype(cls, url):
        surl = saga.url(url)
        surl.query = "" 
        return surl.get_string()
        
    @classmethod
    def __store_entry(cls, entry_url, content):
        entry_url = cls.__get_url(entry_url)
        
        # directory is recursively created
        entry = saga.advert.entry(saga.url(entry_url),
                                           saga.advert.Create | 
                                           saga.advert.CreateParents | saga.advert.ReadWrite)
        entry.store_string(json.dumps(content))
        #logger.debug("Store Advert entry at: " + entry_url 
        #              + " Content: " + str(json.dumps(content)))
        
    @classmethod
    def __retrieve_entry(cls, entry_url):
        entry_url = cls.__get_url(entry_url)
        #logger.debug("Retrieve Advert entry at: " + entry_url)
        # directory is recursively created
        entry = saga.advert.entry(saga.url(entry_url),
                                           saga.advert.Create | 
                                           saga.advert.CreateParents | saga.advert.ReadWrite)
        content = json.loads(entry.retrieve_string())
        #logger.debug("Retrieve Advert entry at: " + entry_url 
        #              + " Content: " + str(json.dumps(content)))
        return content



""" Affinity-aware scheduler that evaluates affinity labels and input/output data flow
    

"""
import random
import pdb
import logging

class Scheduler:
    
    def __init__(self):
        self.pilot_data=[]
        self.pilot_jobs=[]
    
    def set_pilot_data(self, pilot_data):
        """ set resources which are used for scheduling """
        self.pilot_data=pilot_data
        
    
    def set_pilot_jobs(self, pilot_jobs):
        """ set resources which are used for scheduling """
        self.pilot_jobs=pilot_jobs
    
        
    def schedule_pilot_data(self, data_unit_description=None):
        logging.debug("Schedule to PD - # Avail pilots: %d"%len(self.pilot_data))     
        candidate_pilot_data = []  
        if data_unit_description.has_key("affinity_datacenter_label") and data_unit_description.has_key("affinity_machine_label"):
            for i in self.pilot_data: 
                pilot_data_description = i.pilot_data_description
                if data_unit_description["affinity_datacenter_label"] == pilot_data_description["affinity_datacenter_label"]\
                and data_unit_description["affinity_machine_label"] == pilot_data_description["affinity_machine_label"]:
                    candidate_pilot_data.append(i)
        else:
            candidate_pilot_data = self.pilot_data
            
        if len(candidate_pilot_data)>0:
            return random.choice(candidate_pilot_data)
        
        return None
        
        #if len(self.pilot_data)!=0:
        #    return random.choice(self.pilot_data)
        return None
    
    
    def schedule_pilot_job(self, work_unit_description=None):
        """ Enforces affinity description: if no PJ is available with the right
            affinity, WU can't be scheduled.
            
            TODO: incorporate potential data movements to co-locate PD/WU 
        
        """    
        logging.debug("Schedule to PJ - # Avail PJs: %d"%len(self.pilot_jobs))
        candidate_pilot_jobs = []
        if work_unit_description.has_key("affinity_datacenter_label") and work_unit_description.has_key("affinity_machine_label"):
            for i in self.pilot_jobs:
                pilot_job_description = i.pilot_compute_description
                if pilot_job_description["affinity_datacenter_label"] == work_unit_description["affinity_datacenter_label"]\
                and pilot_job_description["affinity_machine_label"] == work_unit_description["affinity_machine_label"]:
                    candidate_pilot_jobs.append(i)
        else:
            candidate_pilot_jobs=self.pilot_jobs
             
        if len(candidate_pilot_jobs)>0:
            return random.choice(candidate_pilot_jobs)
        
        return None
    
    def __check_pilot_data_dependency(self, work_unit_description):
        pilot_data_dependencies = work_unit_description["input_pilot_data"]
        for i in pilot_data_dependencies:
            pd = PilotData.pilot
            ps = i.get_pilot_data()


'''
Encapsulates coordination and communication specifics of bigjob
'''
import logging
import threading
import datetime
import sys
import os
import pickle
import pdb

from bigjob import logger
sys.path.insert(0, (os.path.dirname(os.path.abspath( __file__) ) + "/../ext/redis-2.4.9/"))
from redis import *

if sys.version_info < (2, 5):
    sys.path.append(os.path.dirname( os.path.abspath( __file__) ) + "/../ext/uuid-1.30/")
    sys.stderr.write("Warning: Using unsupported Python version\n")
    
logging.debug(str(sys.path))
import uuid



REDIS_SERVER="localhost"
REDIS_SERVER_PORT=6379
REDIS_URL_SCHEME="redis://"

class bigjob_coordination(object):
    '''
    Encapsulates communication and coordination
    Implementation based on Redis (http://redis.io)
    '''

    def __init__(self, server=REDIS_SERVER, server_port=REDIS_SERVER_PORT, server_connect_url=None,
                 username=None, password=None, dbtype=None, url_prefix=None):
        '''
        Constructor
        '''
        if server_port==None:
            server_port=6379
            
        self.username = None
        self.password = None   
        
        self.address = "%s%s:%i"%(REDIS_URL_SCHEME, server, server_port)
        if server_connect_url!=None:
            self.address=server_connect_url    
            start_index = self.address.find(REDIS_URL_SCHEME)+len(REDIS_URL_SCHEME)
            server_and_port = self.address[start_index:]
            password_end = server_and_port.find("@") 
            # parse out password
            if password_end != -1:
                self.password = server_and_port[:password_end]
                start_index=password_end
                server_and_port= server_and_port[(password_end+1):]
            
            # port and hostname
            if server_and_port.find(":")==-1:
                server=server_and_port
                server_port = REDIS_SERVER_PORT
            else:
                server = server_and_port.split(":")[0]
                server_port = int(server_and_port.split(":")[1])
        else:
            self.password = username
            if self.password != None and self.password!="":
                self.address = "%s%s@%s:%i"%(REDIS_URL_SCHEME, self.password, server, server_port)
        
        logger.debug("Connect to Redis: " + server + " Port: " + str(server_port))
        
        if self.password==None:
            self.redis = Redis(host=server, port=server_port, db=0)
        else:
            self.redis = Redis(host=server, port=server_port, password=self.password, db=0)
        #self.redis_pubsub = self.redis.pubsub() # redis pubsub client       
        self.resource_lock = threading.RLock()
        self.pipe = self.redis.pipeline()
        try:
            self.redis.ping()
        except:
            logger.error("Please start Redis server!")
            raise Exception("Please start Redis server!")
        
    def get_address(self):
        return self.address
    
    #####################################################################################
    # Pilot-Job State
    def set_pilot_state(self, pilot_url, new_state, stopped=False):     
        logger.debug("update state of pilot job to: " + str(new_state))
        self.redis.hmset(pilot_url, {"state":str(new_state), "stopped":str(stopped)})
        
    def get_pilot_state(self, pilot_url):
        state = self.redis.hgetall(pilot_url)
        return state
    
    
    #####################################################################################
    # Pilot-Job State
    def set_pilot_description(self, pilot_url, description):     
        logger.debug("update description of pilot job to: " + str(description))
        self.redis.hmset(pilot_url + ":description", {"description":description})
        
    def get_pilot_description(self, pilot_url):
        description = self.redis.hgetall(pilot_url + ":description")
        return description
    
    
    #def is_pilot_stopped(self,pilot_url):
    #    state = self.redis.hgetall(pilot_url)
    #    if state==None or not state.has_key("stopped"):
    #        return True        
    #    return state["stopped"]
    
    def get_jobs_of_pilot(self, pilot_url):
        """ returns array of job_url that are associated with a pilot """
        return self.redis.keys(pilot_url+":jobs:*") 
    
    def delete_pilot(self, pilot_url):
        items = self.redis.keys(pilot_url+"*")  
        for i in items:        
            self.pipe.delete(i)
        self.pipe.execute()
    
    #####################################################################################
    # Sub-Job State    
    def set_job_state(self, job_url, new_state):
        #self.resource_lock.acquire()        
        logger.debug("set job state to: " + str(new_state))
        self.redis.hset(job_url, "state", str(new_state))
        #self.resource_lock.release()
        
    def get_job_state(self, job_url):
        return self.redis.hget(job_url, "state")      
    
    
    #####################################################################################
    # Sub-Job Description
    def set_job(self, job_url, job_dict):
        self.redis.hmset(job_url, job_dict)
    
    def get_job(self, job_url):
        return self.redis.hgetall(job_url)    
    
    def delete_job(self, job_url):
        self.redis.delete(job_url+"*")
    
    
    #####################################################################################
    # Distributed queue for sub-jobs
    def queue_job(self, pilot_url, job_url):
        """ queue new job to pilot """
        queue_name = pilot_url + ":queue"
        self.redis.set(queue_name + ':last_in', pickle.dumps(datetime.datetime.now()))
        self.redis.lpush(queue_name, job_url)
                
        
    def dequeue_job(self, pilot_url):
        """ deque to new job  of a certain pilot """
        queue_name = pilot_url + ":queue"        
        logger.debug("Dequeue sub-job from: " + queue_name 
                      + " number queued items: " + str(self.redis.llen(queue_name)))
        self.redis.set(queue_name + ':last_out', pickle.dumps(datetime.datetime.now()))
        job_url = self.redis.brpop(queue_name, 120)
        if job_url==None:
            return job_url
        logger.debug("Dequeued: " + str(job_url))
        return job_url[1]
    



#!/usr/bin/env python

#from distutils.core import setup
import os
from setuptools import setup

try:
    import saga
except:
    print "SAGA and SAGA Python Bindings not found: Please install SAGA first (http://saga.cct.lsu.edu)."
    #sys.exit(1)
    
fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'VERSION')
version = open(fn).read().strip()
    
setup(name='BigJob',
      version=version,
      description='SAGA-based Pilot-Job Implementation',
      author='Andre Luckow',
      author_email='aluckow@cct.lsu.edu',
      url='https://github.com/saga-project/BigJob',
      classifiers = ['Development Status :: 4 - Beta',                    
                    'Programming Language :: Python',
                    'Environment :: Console',                    
                    'Topic :: Utilities',
                    ],
      platforms = ('Unix', 'Linux', 'Mac OS'),
      packages=['bigjob', 'bigjob_dynamic', 'coordination', 'pilot', 'pilot.api','pilot.api.compute', 'pilot.api.data', 'pilot.coordination', 
                'pilot.filemanagement', 'pilot.impl', 'pilot.scheduler', 'examples', 'api', 'bootstrap'],
      #data_files=['bigjob_agent.conf', 
      #            'bigjob.conf', "VERSION", "README.md"],
      data_files = [( '', ['bigjob_agent.conf', 'bigjob_agent.conf']),  
                    ('', ['bigjob.conf', 'bigjob.conf']), 
                    ('', ['README.md', 'README.md']), 
                    ('', ['VERSION', 'VERSION'])],
#      package_data = {
#        '': ['bigjob_agent.conf', 'bigjob_agent.conf'],
#        '': ['bigjob.conf', 'bigjob.conf'],
#        '': ['README', 'README'],
#        '': ['VERSION', 'VERSION']
#      },
      install_requires=['paramiko-on-pypi', 'uuid', 'threadpool', 'virtualenv', 'redis==2.2.4', 'bliss'],
      entry_points = {
        'console_scripts': [
            'test-bigjob = examples.example_local_single:main',
            'test-bigjob-dynamic = examples.example_manyjob_local:main'            
        ]
        }
)



from pilot.api.compute.api import PilotCompute as MyPilotCompute
from pilot.api.compute.api import PilotComputeService as MyPilotComputeService
from pilot.api.compute.api import PilotComputeDescription as MyPilotComputeDescription
from pilot.api.compute.api import ComputeUnit as MyComputeUnit
from pilot.api.compute.api import ComputeUnitService as MyComputeUnitService
from pilot.api.compute.api import ComputeUnitDescription as MyComputeUnitDescription

from pilot.api.compute.api import State as MyState

from pilot.api.data.api import PilotDataDescription as MyPilotDataDescription
from pilot.api.data.api import PilotData as MyPilotData
from pilot.api.data.api import PilotDataService as MyPilotDataService
from pilot.api.data.api import DataUnitService as MyDataUnitService
from pilot.api.data.api import DataUnit as MyDataUnit
from pilot.api.data.api import DataUnitDescription as MyDataUnitDescription

from pilot.api.api import ComputeDataDescription as MyComputeDataDescription
from pilot.api.api import ComputeDataService as MyComputeDataService

###############################################################################
# Common state model

class State(MyState):
    pass


###############################################################################
# Compute only

class PilotCompute(MyPilotCompute):
    pass


class PilotComputeService(MyPilotComputeService):
    pass


class PilotComputeDescription(MyPilotComputeDescription):
    pass


class ComputeUnit(MyComputeUnit):
    pass


class ComputeUnitService(MyComputeUnit):
    pass


class ComputeUnitDescription(MyComputeUnitDescription):
    pass


###############################################################################
# Data only

class PilotData(MyPilotData):
    pass


class PilotDataService(MyPilotDataService):
    pass


class PilotDataDescription(MyPilotDataDescription):
    pass


class DataUnit(MyDataUnit):
    pass


class DataUnitService(MyDataUnitService):
    pass


class DataUnitDescription(MyDataUnitDescription):
    pass



###############################################################################
# Compute and Data

class ComputeDataDescription(MyComputeDataDescription):
    pass


class ComputeDataService(MyComputeDataService):
    pass
